#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include "file_handle.h"
#include "treedef.h"



/* local */



int file_length (const char *filename)
{
  struct stat stbuf;

  if (stat(filename,&stbuf) == -1)
    {
      perror("failed to 'stat' for file length");
      exit(1);
    }

  return(stbuf.st_size);
}



void rec_save_spst (FILE *fp, int absize, spst_type T)
{
  int i;
  
  fprintf(fp,"%s ",T->label);
  for(i=0; i<=absize; i++)
      fprintf(fp,"%d ",T->dist[i]);
  for(i=0; i<absize; i++)
  {
    fprintf(fp,"%d ",T->sonsp[i] != NULL);             
    if ((T->sonsp)[i] != NULL)
	  rec_save_spst(fp, absize, (T->sonsp)[i]);
  }
}



void rec_load_spst (FILE *fp, int absize, spst_type *Tp)

{
  int i;
  int *N;
  int son;
  char *label;

  if ((label = (char *) malloc ((absize+1)*sizeof(char))) == NULL)
    {
      perror("failed to allocate a tree node probability buffer\n");
      exit(1);
    }
  if ((N = (int *) malloc ((absize+1)*sizeof(int))) == NULL)
    {
      perror("failed to allocate a tree node probability buffer\n");
      exit(1);
    }
  fscanf(fp,"%s ",label);
  for (i=0;i<=absize;i++)
     fscanf(fp,"%d ",N+i);
  (*Tp) = make_spst_node(absize,label,N);
  free(label);
  free(N);
  for(i=0; i<absize; i++)
    {
      fscanf(fp,"%d ",&son);
      if (son)
        rec_load_spst(fp, absize, (*Tp)->sonsp+i);
    }
}




/* public */



void read_alphabet (int *absizep, char **ABp, const char *filename)
{
  FILE *fp;

  if ((fp = fopen(filename, "r")) == NULL)
      {
	perror("failed to open <ab-file> for read");
	exit(1);
      }
  if ((*ABp = (char *) malloc (file_length(filename))) == NULL)
    {
      perror("failed to allocate AB\n");
      exit(1);
    }
  fscanf(fp,"%s",*ABp);
  *absizep = strlen(*ABp);
  fclose(fp);
  return;
}


void read_fasta_set (char **data_setp, int *mp, char ***ds_ptrp, \
		     int **lip, const char *AB, const char *filename, \
		     const char *indfile, int *tez, int *trz, int *trl)
{
  FILE *fp;
  FILE *ifp;
  char cbuf[STRINGLEN_MAX+1];
  int cbuflen;
  int i;
  int ind;
  int loc;
  int indexed, flag, state, incoming;

  if ((fp = fopen(filename, "r")) == NULL)
      {
	perror("failed to open <strings_file> for read");
	exit(1);
      }

  indexed = (*indfile != '@');

  if ((indexed) && ((ifp = fopen(indfile, "r")) == NULL))
      {
	perror("failed to open <index_file> for read");
	exit(1);
      }

  /* 1st pass - count size to allocate, verify content */

  *tez = 0;
  *mp = 0;
  loc = 0;  /* calcs li[0] = total string length */
  flag = 1;
  state = -1;

  while (flag)
    {

      do
	{
	  if (fgets(cbuf, STRINGLEN_MAX, fp) == NULL)
	    {
	      incoming = 0; /* EOF or error :-( */
	      cbuflen = -1;
	    }
	  else
	    {
	      if ((cbuflen = strlen(cbuf)) >= STRINGLEN_MAX-1) /* max poss */
		{
		  fprintf(stderr, "single LINE size seems to exceeds STRINGLEN_MAX = %d\n (you may update and re-compile)\n", STRINGLEN_MAX);
		  exit(1);
		}
	      if (cbuf[cbuflen-1] == '\n')
		{
		  cbuf[--cbuflen] = '\0';
		}
	      if (cbuflen > 0)
		if (*cbuf == '>')
		  incoming = 1; /* comment line */
		else
		  {
		    incoming = 2; /* data line */
		    if (strspn(cbuf,AB)!=cbuflen)
		      {
			fprintf(stderr,"found a rejected data string:\n%s\n", cbuf);
			exit(1);
		      }
		  }
	    }
	}
      while (cbuflen == 0);
	
      switch (state)
	{
	case -1: /* start of file */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: empty string_file\n");
	      exit(1);
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      fprintf(stderr,"Read fasta: data start without comment line\n");
	      exit(1);
	      break;
	    default:
	      fprintf(stderr,"c-code bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  break;

	case  0: /* after comment */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: comment line without data\n");
	      exit(1);
	      break;
	    case 1:
	      fprintf(stderr,"Read fasta: two consecutive comment lines\n");
	      exit(1);
	      break;
	    case 2:
	      if ((indexed) && ((fscanf(ifp,"%d",&ind)) == EOF))
		{
		  fprintf(stderr,"index_file is shorter than string_file\n");
		  exit(1);
		}
	      if ((indexed) && (ind))
		{
		  (*tez)++;
		  state = 2;
		}
	      else /* non-indexed || ind == 0 */
		{
		  (*mp)++;
		  loc += cbuflen;
		  state = 1;
		}
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;
	  
	case  1: /* in goodata */
	  switch(incoming)
	    {
	    case 0:
	      flag = 0;
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      loc += cbuflen;
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;

	case  2: /* in badata */
	  switch(incoming)
	    {
	    case 0:
	      flag = 0;
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;
	  
	default:
	  fprintf(stderr,"Program bug: impossible state in read fasta\n");
	  exit(1);
	  break;
	}
    }

  *trz = *mp;
  *trl = loc;
    
  rewind(fp);
  if (indexed)
    rewind(ifp);

  if ((*data_setp = (char *) malloc (loc+(*mp)+1)) == NULL)
    { /* format: \0 string \0 ... \0 string \0 */
      perror("failed to allocate data set storage char array\n");
      exit(1);
    }
  if (((*ds_ptrp = (char **) malloc (((*mp)+1)*sizeof(char *))) == NULL) ||
      ((*lip     = (int *)   malloc (((*mp)+1)*sizeof(int   ))) == NULL))
    {
      perror("failed to allocate data set storage related variables\n");
      exit(1);
    }
  (*lip)[0] = loc;
  (*ds_ptrp)[0] = NULL;    /* we don't use it */
  (*data_setp)[0] = '\0';  /* we need it for efficient string count */
  
  /* 2nd pass - read into memory */

  i = 0; /* string number, like *mp above */
  loc = 1; /* now calcs relative pos within *data_setp, [0] contains '\0' */
  flag = 1;
  state = -1;

  while (flag)
    {

      do
	{
	  if (fgets(cbuf, STRINGLEN_MAX, fp) == NULL)
	    {
	      incoming = 0; /* EOF or error :-( */
	      cbuflen = -1;
	    }
	  else
	    {
	      if ((cbuflen = strlen(cbuf)) >= STRINGLEN_MAX-1) /* max poss */
		{
		  fprintf(stderr, "single string size roughly exceeds STRINGLEN_MAX = %d (may update and re-compile)\n", STRINGLEN_MAX);
		  exit(1);
		}
	      if (cbuf[cbuflen-1] == '\n')
		{
		  cbuf[--cbuflen] = '\0';
		}
	      if (cbuflen > 0)
		if (*cbuf == '>')
		  incoming = 1; /* comment line */
		else
		  {
		    incoming = 2; /* data line */
		    if (strspn(cbuf,AB)!=cbuflen)
		      {
			fprintf(stderr,"found a rejected data string:\n%s\n", cbuf);
			exit(1);
		      }
		  }
	    }
	}
      while (cbuflen == 0);
      
      switch (state)
	{
	case -1: /* start of file */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: empty string_file\n");
	      exit(1);
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      fprintf(stderr,"Read fasta: data start without comment line\n");
	      exit(1);
	      break;
	    default:
	      fprintf(stderr,"c-code bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  break;

	case  0: /* after comment */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: comment line without data\n");
	      exit(1);
	      break;
	    case 1:
	      fprintf(stderr,"Read fasta: two consecutive comment lines\n");
	      exit(1);
	      break;
	    case 2:
	      if ((indexed) && ((fscanf(ifp,"%d",&ind)) == EOF))
		{
		  fprintf(stderr,"index_file is shorter then string_file\n");
		  exit(1);
		}
	      if ((indexed) && (ind))
		state = 2;
	      else /* non-indexed || ind == 0 */
		{
		  (*lip)[++i] = cbuflen;
		  strcpy((*data_setp)+loc,cbuf); /* trailing \0 is copied */
		  (*ds_ptrp)[i] = (*data_setp)+loc;
		  loc += cbuflen; /* trailing \0 to be trampled unless last */
		  state = 1;
		}
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;
	  
	case  1: /* in goodata */
	  switch(incoming)
	    {
	    case 0:
	      loc++; /* last trailing \0 */
	      flag = 0;
	      break;
	    case 1:
	      loc++; /* string trailing \0 */
	      state = 0;
	      break;
	    case 2:
	      (*lip)[i] += cbuflen;
	      strcpy((*data_setp)+loc,cbuf); /* trailing \0 is copied */
	      loc += cbuflen; /* trailing \0 to be trampled unless last */
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;

	case  2: /* in badata */
	  switch(incoming)
	    {
	    case 0:
	      flag = 0;
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;
	  
	default:
	  fprintf(stderr,"Program bug: impossible state in read fasta\n");
	  exit(1);
	  break;
	}
    }

  fclose(fp);
  if (indexed)
    fclose(ifp);

  return;

}



void predict_fasta_set (double gam_min, int n_mean, const char *out_file, \
			spst_type T, char *AB, const char *filename, \
			const char *indfile, int *tez, int *trz, int *trl)
{
  FILE *fp;
  FILE *ifp;
  FILE *sfp;
  char cbuf[STRINGLEN_MAX+1];
  int cbuflen;
  int ind;
  int loc;
  int indexed, flag, state, incoming;
  char *entryp;
  int maxloc;
  char D[1001]; /* first 1000 chars of description line are echoed */
  
  if ((fp = fopen(filename, "r")) == NULL)
    {
      perror("failed to open <strings_file> for read");
      exit(1);
    }
  
  indexed = (*indfile != '@');
  
  if ((indexed) && ((ifp = fopen(indfile, "r")) == NULL))
    {
      perror("failed to open <index_file> for read");
      exit(1);
    }
  
  /* 1st pass - count size to allocate, verify content */
  
  *trl = 0;    /* total predicting length */
  *tez = 0;    /* ignored size */
  *trz = 0;    /* predicting size */
  maxloc = 0;  /* calcs max entry length */
  flag = 1;
  state = -1;
  
  while (flag)
    {
      
      do
	{
	  if (fgets(cbuf, STRINGLEN_MAX, fp) == NULL)
	    {
	      incoming = 0; /* EOF or error :-( */
	      cbuflen = -1;
	    }
	  else
	    {
	      if ((cbuflen = strlen(cbuf)) >= STRINGLEN_MAX-1) /* max poss */
		{
		  fprintf(stderr, "single LINE size seems to exceeds STRINGLEN_MAX = %d\n (you may update and re-compile)\n", STRINGLEN_MAX);
		  exit(1);
		}
	      if (cbuf[cbuflen-1] == '\n')
		{
		  cbuf[--cbuflen] = '\0';
		}
	      if (cbuflen > 0)
		if (*cbuf == '>')
		  incoming = 1; /* comment line */
		else
		  {
		    incoming = 2; /* data line */
		    if (strspn(cbuf,AB)!=cbuflen)
		      {
			fprintf(stderr,"found a rejected data string:\n%s\n", cbuf);
			exit(1);
		      }
		  }
	    }
	}
      while (cbuflen == 0);
	
      switch (state)
	{
	case -1: /* start of file */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: empty string_file\n");
	      exit(1);
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      fprintf(stderr,"Read fasta: data start without comment line\n");
	      exit(1);
	      break;
	    default:
	      fprintf(stderr,"c-code bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  break;

	case  0: /* after comment */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: comment line without data\n");
	      exit(1);
	      break;
	    case 1:
	      fprintf(stderr,"Read fasta: two consecutive comment lines\n");
	      exit(1);
	      break;
	    case 2:
	      if ((indexed) && ((fscanf(ifp,"%d",&ind)) == EOF))
		{
		  fprintf(stderr,"index_file is shorter then string_file\n");
		  exit(1);
		}
	      if ((indexed) && (ind))
		{
		  (*tez)++;
		  state = 2;
		}
	      else /* non-indexed || ind == 0 */
		{
		  (*trz)++;
		  (*trl) += cbuflen;
		  loc = cbuflen; /* restarted */
		  state = 1;
		}
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;
	  
	case  1: /* in goodata */
	  switch(incoming)
	    {
	    case 0:
	      if (loc > maxloc)
		maxloc = loc;
	      flag = 0;
	      break;
	    case 1:
	      if (loc > maxloc)
		maxloc = loc;
	      state = 0;
	      break;
	    case 2:
	      (*trl) += cbuflen;
	      loc += cbuflen;
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }

	  break;

	case  2: /* in badata */
	  switch(incoming)
	    {
	    case 0:
	      flag = 0;
	      break;
	    case 1:
	      state = 0;
	      break;
	    case 2:
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  
	  break;
	  
	default:
	  fprintf(stderr,"Program bug: impossible state in read fasta\n");
	  exit(1);
	  break;
	}
    }
  
  rewind(fp);
  if (indexed)
    rewind(ifp);

  if ((entryp = (char *) malloc (maxloc+1+1)) == NULL)
    { /* format: \0 string \0 */
      perror("failed to allocate max entry storage char array\n");
      exit(1);
    }
  entryp[0] = '\0';  /* we need it for efficient string count */

  /* 2nd pass - read entry into memory, predict, repeat */

    
  if ((sfp = fopen(out_file, "w")) == NULL)
    {
      perror("failed to open <out_file> for write");
      exit(1);
    }
  
  flag = 1;
  state = -1;
  D[1000] = '\0';
  
  while (flag)
    {

      do
	{
	  if (fgets(cbuf, STRINGLEN_MAX, fp) == NULL)
	    {
	      incoming = 0; /* EOF or error :-( */
	      cbuflen = -1;
	    }
	  else
	    {
	      if ((cbuflen = strlen(cbuf)) >= STRINGLEN_MAX-1) /* max poss */
		{
		  fprintf(stderr, "single string size roughly exceeds STRINGLEN_MAX = %d (may update and re-compile)\n", STRINGLEN_MAX);
	          exit(1);
		}
	      if (cbuf[cbuflen-1] == '\n')
		{
		  cbuf[--cbuflen] = '\0';
		}
	      if (cbuflen > 0)
		if (*cbuf == '>')
		  {
		    if (cbuflen > 1000)
		      {
			fprintf(stderr, "single comment line size roughly exceeds 1000 characters\n");
			exit(1);
		      }
		    incoming = 1; /* comment line */
		  }
		else
		  {
		    incoming = 2; /* data line */
		    if (strspn(cbuf,AB)!=cbuflen)
		      {
			fprintf(stderr,"found a rejected data string:\n%s\n", cbuf);
			exit(1);
		      }
		  }
	    }
	}
      while (cbuflen == 0);
      
      switch (state)
	{
	case -1: /* start of file */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: empty string_file\n");
	      exit(1);
	      break;
	    case 1:
	      strncpy(D, cbuf, 1000);
	      state = 0;
	      break;
	    case 2:
	      fprintf(stderr,"Read fasta: data start without comment line\n");
	      exit(1);
	      break;
	    default:
	      fprintf(stderr,"c-code bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  break;

	case  0: /* after comment */
	  switch(incoming)
	    {
	    case 0:
	      fprintf(stderr,"Read fasta: comment line without data\n");
	      exit(1);
	      break;
	    case 1:
	      fprintf(stderr,"Read fasta: two consecutive comment lines\n");
	      exit(1);
	      break;
	    case 2:
	      if ((indexed) && ((fscanf(ifp,"%d",&ind)) == EOF))
		{
		  fprintf(stderr,"index_file is shorter then string_file\n");
		  exit(1);
		}
	      if ((indexed) && (ind))
		state = 2;
	      else /* non-indexed || ind == 0 */
		{
		  loc = 1; 
		  /* restarted, calcs relative pos within entryp, [0] contains '\0' */
		  strcpy(entryp+loc,cbuf); /* trailing \0 is copied */
		  loc += cbuflen; 
		  /* restarted, trailing \0 to be trampled unless last */
		  state = 1;
		}
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  
	  break;
	  
	case  1: /* in goodata */
	  switch(incoming)
	    {
	    case 0:
	      predict_entry (strlen(AB),AB,T,entryp+1,gam_min,sfp,D,n_mean);
	      flag = 0;
	      break;
	    case 1:
	      predict_entry (strlen(AB),AB,T,entryp+1,gam_min,sfp,D,n_mean);
	      strncpy(D, cbuf, 1000);
	      state = 0;
	      break;
	    case 2:
	      strcpy(entryp+loc,cbuf); /* trailing \0 is copied */
	      loc += cbuflen; /* trailing \0 to be trampled unless last */
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  
	  break;
	  
	case  2: /* in badata */
	  switch(incoming)
	    {
	    case 0:
	      flag = 0;
	      break;
	    case 1:
	      strncpy(D, cbuf, 1000);
	      state = 0;
	      break;
	    case 2:
	      break;
	    default:
	      fprintf(stderr,"Program bug: impossible state in read fasta\n");
	      exit(1);
	      break;
	    }
	  
	  break;
	  
	default:
	  fprintf(stderr,"Program bug: impossible state in read fasta\n");
	  exit(1);
	  break;
	}
    }
  
  fclose(fp);
  if (indexed)
    fclose(ifp);
  fclose(sfp);
  
  return;
  
}



void save_spst (char *outfile, char *AB, int absize, int L_max, spst_type T)
/* clobbers existing if found */
{
  FILE *fp;

  if ((fp = fopen(outfile, "w")) == NULL)
      {
	perror("failed to open <output_file> for write");
	exit(1);
      }
  fprintf(fp, "%d %s %d ", absize, AB, L_max);
  rec_save_spst (fp, absize, T);
  fclose(fp);
  return;
}



void load_spst (char *spstfile, char **ABp, int *absizep, int *L_maxp, 
	       spst_type *Tp)
{
  FILE *fp;

  if ((fp = fopen(spstfile, "r")) == NULL)
      {
	perror("failed to open <tree-file> for read");
	exit(1);
      }
  fscanf(fp, "%d ", absizep);
  if ((*ABp = (char *) malloc ((*absizep)+1)) == NULL)
    {
      perror("failed to allocate AB buffer\n");
      exit(1);
    }
  fscanf(fp,"%s %d ",*ABp, L_maxp);
  rec_load_spst (fp, *absizep, Tp);
  fclose(fp);
  return;
}

