#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "learn_alg.h"
#include "queue.h"
#include "pattern_match.h"



/* local */



double distance (int *N, int *M, int dim)
     /* N e M sao vetores com dim+1 componentes */
{
  int i;
  double loglike,log1,log2,log3;
  loglike = 0.0;
  for(i=0; i<dim; i++)
    {
      if ((N[i]+M[i]) > 0.0)
	log1 = (N[i]+M[i])*(log((double)(N[i]+M[i])/(double)(N[dim]+M[dim])));
      else
	log1 = 0.0;
      if (N[i] > 0.0)
	log2 =  N[i]*log((double)N[i]/(double)N[dim]);
      else 
	log2 = 0.0;
      if (M[i] > 0.0)
	log3 =  M[i]*log((double)M[i]/(double)M[dim]);
      else 
	log3 = 0.0;
      loglike += log1 - log2 - log3;
    }
  return(loglike);
}



int merge_nodes(spst_type Tp, double r, int nodes, int absize)
     // nodes > 1
{
  int i,j,min_i,min_j;
  double loglike,logmax;
  spst_type T1,T2;
  if (nodes > 1)
    {
      T1 = (Tp->sonsp)[0];
      T2 = (Tp->sonsp)[1];
      min_i = 0;
      min_j = 1;
      logmax = distance(T1->dist,T2->dist,absize);
    }
  else
    return (0);
  for (i=0;i<(nodes-1);i++)
    for(j=i+1;j<nodes;j++)
      {
	T1 = (Tp->sonsp)[i];
	T2 = (Tp->sonsp)[j];
	loglike = distance(T1->dist,T2->dist,absize);
	if (loglike > logmax)
	  {
	    logmax = loglike;
	    min_i = i;
	    min_j = j;
	  }
      }
  T1 = (Tp->sonsp)[min_i];
  T2 = (Tp->sonsp)[min_j];
  if (-2*logmax < r)
    {
      strcat(T1->label,T2->label);
      for(i=0;i<=absize;i++)
	(T1->dist)[i] += (T2->dist)[i];
      for(j=min_j;j<(nodes-1);j++)
	(Tp->sonsp)[j] = (Tp->sonsp)[j+1];
      (Tp->sonsp)[nodes-1] = NULL;
      free_spst_node(T2,absize);
      return (1);
    }
  else 
    return(0);
}



void merge_spst_nodes(spst_type Tp, double r, int nodes, int absize)
{
  while (merge_nodes(Tp, r, nodes, absize))
    nodes--;
}



/* public */



spst_type grow_tree(char *AB, int absize, int m, char **ts_ptr, int trl)
{

  spst_type T, Tp, Ts;
  q_type Q; /* used as a LIFO stack */
  int i, j, k, len, count;
  int key, keym, depth_s;
  char c;
  char *s;
  char **vect;
  char label[2];
  int *N, **M; 
  double r;
  
  if ((N = (int *) malloc ((absize+1)*sizeof(int))) == NULL)
    {
      perror("grow_tree(): learn_alg.c: failed to allocate N\n");
      exit(1);
    }
  if ((M = (int **) malloc (absize*sizeof(int *))) == NULL)
    {
      perror("grow_tree(): learn_alg.c: failed to allocate M\n");
      exit(1);
    }
  for (i=0;i<absize;i++)
    if ((M[i] = (int *) malloc (absize*sizeof(int))) == NULL)
      {
	perror("grow_tree(): learn_alg.c: failed to allocate M\n");
	exit(1);
      }
  if ((vect = (char **) malloc (L_max*sizeof(char *))) == NULL)
    {
      perror("grow_tree(): learn_alg.c: failed to allocate vect\n");
      exit(1);
    } 
  if ((s = (char *) malloc(L_max*(absize+1))) == NULL)
    {
      fprintf (stderr,"grow_tree(): learn_alg.c: failed to allocate sequence seq\n");
      exit(1);
    } 
  
  vect[0] = s;
  vect[0][0] = '\0';
  for (i=1;i<L_max;i++)
    {
      vect[i] = vect[i-1] + absize + 1;
      vect[i][0] = '\0';
    }
  
  r = r_max; 

    /* inicialization */

  Q = make_empty_queue();
  calc_dist(AB, absize, N, M[0], vect, 0, L_max, m, ts_ptr);
  T = make_root_node(absize,N);
  vect[L_max-1][1] = '\0';
  key = 1;
  for (i=1; i<=absize; i++)
    {
      vect[L_max-1][0] = AB[absize-i];
      if (M[0][absize-i] >= N_min)
	enqueue(&Q, key, 1, L_max, vect, absize);
    }
  

  /* tree grow */
  
  while (not_empty_queue(Q))
    {
      i = 0;
      dequeue(&Q, &keym, &depth_s, L_max, vect, &c);
      calc_dist(AB, absize, N, M[ABind(AB,c)], vect, depth_s, L_max, m, ts_ptr);
      label[1] = '\0';
      label[0] = c;
      Tp = traverse_spst(absize, T, vect, depth_s, L_max);
      (Tp->sonsp)[i] = make_spst_node(absize,label,N);
      while (not_empty_queue(Q) && (Q->key == keym))
	{
	  i++;
	  dequeue(&Q, &keym, &depth_s, L_max, vect, &c);
	  calc_dist(AB, absize, N, M[ABind(AB,c)], vect, depth_s, L_max, m, ts_ptr);
	  label[0] = c;
	  (Tp->sonsp)[i] = make_spst_node(absize,label,N);
	}
      i++;
      merge_spst_nodes(Tp,r,i,absize); // substituir r por r/depth_s
      if (depth_s < L_max)
	{
	  i = 0;
	  while (i<absize && (Tp->sonsp)[i] != NULL)
	    {
	      ++key;
	      Ts = (Tp->sonsp)[i];
	      vect[L_max-depth_s-1][1] = '\0';
	      for (j=1;j<=absize;j++)
		{
		  count = 0;
		  for (k=0;k<strlen(Ts->label);k++)
		    count += M[ABind(AB,Ts->label[k])][absize-j];
		  if (count >= N_min)
		    {
		      strcpy(vect[L_max-depth_s],Ts->label);
		      vect[L_max-depth_s-1][0] = AB[absize-j];
		      enqueue(&Q, key, depth_s+1, L_max, vect, absize);
		    }
		}
	      ++i;
	    }
	}
    }
  return(T);
}
