/* Predicting from an existing SPST */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "file_handle.h"


int main (int argc, char **argv)
{


  int absize, test_size, n_mean;
  int train_size, train_len, max_L;
  char *AB, *spst_file, *strings_file, *index_file;
  char *out_file;
  double gamma_min, smooth_p;
  spst_type T;
  FILE *fp;
  
  
  switch (argc)
    {
    case 1:  /* short reminder */
      system("clear");
      printf(" (clear screen)\n");
      fprintf(stderr,"GENERAL call format:\n\nPrediction using a given SPST:\n\n");
      fprintf(stderr," %s <spst_file> <strings_file> <index_file/@> <gamma_min> ...\n",argv[0]);
      fprintf(stderr," ... <n_mean> <out_file> \n\n\n");
      fprintf(stderr,"<strings_file> should be in FASTA format.\n");
      fprintf(stderr,"<index_file>   entries: 0 = read, other = ignore (no index => read all)\n");
      fprintf(stderr,"<n_mean>       if 0 general SPST, if > 0 F-SPST\n");
      fprintf(stderr,"<out_file>     {sum_log10like comment} x no_of_strings\n\n");
      exit(1);
        
    case 7:
      spst_file = argv[1];
      strings_file = argv[2];
      index_file = argv[3];
      gamma_min = atof(argv[4]);
      n_mean = atoi(argv[5]);
      out_file = argv[6];
      break;
      
    default:
      fprintf(stderr,"Syntax Error: type %s for a short reminder\n",argv[0]);
      exit(1);

    }
  

  /* predict */
 
  load_spst (spst_file, &AB, &absize, &max_L, &T);

  /* now we can check whether smoothing is possible */
  smooth_p = 1 - absize*gamma_min;
  if (smooth_p <= 0)
    {
      fprintf(stderr, "Cannot smooth. Decrease gamma_min.\n\n");
      exit(1);
    }
  if (n_mean == 0)
    n_mean = STRINGLEN_MAX;
  predict_fasta_set(gamma_min, n_mean, out_file, T, AB, strings_file, index_file, 
		    &test_size, &train_size, &train_len);
  return(0);
}    

