#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "treedef.h"
#include "pattern_match.h"



/* local */



int Pind (int absize, spst_type nodep, char c)
{
  int i;
  
  i = 0;
  while (i<absize && (nodep->sonsp)[i] != NULL)
    if (strchr((nodep->sonsp)[i]->label,c) != NULL)
      return i;
    else
      ++i;
  return -1;     
}



/* public */



spst_type make_root_node(int absize, int *N)
{
  spst_node *nodep;
  int i;
    
  if ((nodep = (spst_node *) malloc (sizeof(spst_node))) == NULL)
    {
      fprintf (stderr,"failed to allocate root node\n");
      exit(1);
    }
  if ((nodep->label = (char *) malloc (5*sizeof(char))) == NULL)
    {
      fprintf (stderr,"failed to allocate root node->label\n");
      exit(1);
    }
  strcpy(nodep->label,"root");
  if ((nodep->dist = (int *) malloc ((absize+1)*sizeof(int))) == NULL)
    {
      fprintf (stderr,"failed to allocate root node->dist\n");
      exit(1);
    }
  for (i=0; i<=absize; i++)
    (nodep->dist)[i] = N[i];
  if ((nodep->sonsp = (spst_node **) malloc (absize*sizeof(spst_node *))) == NULL)
    {
      fprintf (stderr,"failed to allocate root node->sonsp\n");
      exit(1);
    }
  for (i=0; i<absize; i++)
    (nodep->sonsp)[i] = NULL;
  return(nodep);
}


spst_type make_spst_node (int absize, char *s, int *N)
{
  spst_node *nodep;
  int i;
  int l;
  
  if ((nodep = (spst_node *) malloc (sizeof(spst_node))) == NULL)
    {
      fprintf (stderr,"failed to allocate spst node\n");
      exit(1);
    }
  if ((nodep->label = (char *) malloc ((absize+1)*sizeof(char))) == NULL)
    {
      fprintf (stderr,"failed to allocate spst node->label\n");
      exit(1);
    }
  strcpy(nodep->label,s);
  if ((nodep->dist = (int *) malloc ((absize+1)*sizeof(int))) == NULL)
    {
      fprintf (stderr,"failed to allocate spst node->dist\n");
      exit(1);
    }
  for (i=0; i<(absize+1); i++)
    (nodep->dist)[i] = N[i];
  if ((nodep->sonsp = (spst_node **) malloc (absize*sizeof(spst_node *))) == NULL)
    {
      fprintf (stderr,"failed to allocate spst node->sonsp\n");
      exit(1);
    }
  for (i=0; i<absize; i++)
    (nodep->sonsp)[i] = NULL;
  return(nodep);
}



void free_spst_node (spst_type Tp, int absize)
{
  int i;
  if (Tp != NULL)
    {
      if (Tp->label != NULL)
	free(Tp->label);
      if (Tp->dist != NULL)
	free(Tp->dist);
      if (Tp->sonsp != NULL)
	free(Tp->sonsp);
      free (Tp);
      Tp = NULL;
    }
  return;
}



spst_type traverse_spst(int absize, spst_type T, char **vect, int depth_s, int L)
     /* output:  pointer to leaf of traverse */
     /* handles empty string correctly.
	reads string right to left, suffix-wise! */
{
  int i, ind, cp;       
  spst_node *nextp;
  
  if (depth_s == 1)
    for (i=0;i<absize;i++)
      if ((T->sonsp)[i] != NULL)
	{
	  fprintf (stderr,"traverse_spst(): treedef.c: string ended but tree continues\n");
	  exit(1);
	}
      else
	return T;
  else 
    {
      cp = 1;
      ind = Pind(absize,T,vect[L-cp][0]);
      if (ind >= 0)
	nextp = (T->sonsp)[ind];
      else
	{
	  fprintf (stderr,"traverse_spst(): treedef.c: bad tree\n");
	  exit(1);
	}
      while (nextp != NULL) /* the tree continues */
	{
	  cp++;
	  T = nextp;
	  if (cp == depth_s) /* the string ends */
	    break;
	  ind = Pind(absize,T,vect[L-cp][0]);
	  if (ind >= 0)  
	    nextp = (T->sonsp)[ind];
	  else 
	    {
	      fprintf (stderr,"traverse_spst: treedef.c: bad tree\n");
	      exit(1);
	    }
	}
    }
  return T;
}



spst_type search_spst(int absize, spst_type T, char *s, int *lri)
{
  int ind;       
  spst_node *nextp;
  
  (*lri) = strlen(s);
  if (*lri == 0)
    return T;
  else
    {
      ind = Pind(absize,T,s[(*lri)-1]);
      if (ind >= 0)
	nextp = (T->sonsp)[ind];
    }
  while (nextp != NULL) /* the tree continues */
    {
      T = nextp;
      (*lri)--;
      ind = Pind(absize,T,s[(*lri)-1]);
      if (*lri == 0) /* the string ends */
	break;
      if (ind >= 0)  
        nextp = (T->sonsp)[ind];
      else 
        nextp = NULL;  
    }
  return T;
}




void predict_entry (int absize, char *AB, spst_type T, char *S, double gam_min, 
                    FILE *sfp, char *D, int n_mean)
{
  char c;
  int i,j;
  int lri;
  int lenS;
  spst_node *leafp;
  double loglike, persymbol, weight, max;

  lenS = strlen(S);
  weight =  1 - absize * gam_min;
  loglike = 0.0;
  if ((n_mean == 0) || (lenS < n_mean))
    {
      for (i=0; i<lenS; i++)
	{
	  c = S[i];
	  S[i] = '\0';
	  leafp = search_spst(absize, T, S, &lri);
	  persymbol = ((double)(leafp->dist[ABind(AB,c)]))/((double)(leafp->dist[absize])); /* the prob */
	  persymbol = persymbol * weight + gam_min; /*smoothed probability */
	  loglike += log10(persymbol);
	  S[i] = c;
	}
      loglike = loglike/(double)lenS;
      fprintf(sfp, "%1.15f\t%-s <\n", loglike, D);
      return; 
    }
  for (i=0;i<n_mean;i++)
    {
      c = S[i];
      S[i] = '\0';
      leafp = search_spst(absize, T, S, &lri);
      persymbol = ((double)(leafp->dist[ABind(AB,c)]))/((double)(leafp->dist[absize])); /* the prob */
      persymbol = persymbol * weight + gam_min; /*smoothen probability */
      loglike += log10(persymbol);
      S[i] = c;
    }
  max = loglike;
  for (i=n_mean;i<lenS;i++)
    {
      c = S[i];
      S[i] = '\0';
      leafp = search_spst(absize, T, S, &lri);
      persymbol = ((double)leafp->dist[ABind(AB,c)])/((double)(leafp->dist[absize])); /* the prob */
      persymbol = persymbol * weight + gam_min; /*smoothen probability */
      loglike += log10(persymbol);
      S[i] = c;
      c = S[i-n_mean];
      S[i-n_mean] = '\0';
      leafp = search_spst(absize, T, S, &lri);
      persymbol = ((double)leafp->dist[ABind(AB,c)])/((double)(leafp->dist[absize])); /* the prob */
      persymbol = persymbol * weight + gam_min; /*smoothen probability */
      loglike -= log10(persymbol);
      S[i-n_mean] = c;
      if (loglike > max)
	max = loglike;
    }
  max = max/(double) n_mean;
  fprintf(sfp, "%1.15f\t%-s <\n", max, D);
  return;
}

    



